> module Yahc.Interaction where
> import Yahc.Derivation (Der, step, conclusion, mkder, addStep, undoStep, result, steps, lhsDer, stepsub)
> import Yahc.Expr (QExpr, ConName, unlift, equiv)
> import Yahc.Rules (parseRule, showRules, Axiom(prop), rules, axiom, definition, name)
> import Yahc.Pretty (Pretty, render, pretty)
> import Yahc.Subst (Substitution)
> import Yahc.Parser (parseExp, parseConst)
> import Yahc.Command 
> import Yahc.Error
> import Yahc.Utils
> import Yahc.Checker

> import Control.Monad
> import Control.Monad.State
> import Control.Arrow
> import Data.Maybe (maybe, isNothing)
> import Data.Char (isSpace, toUpper)
> import System.Exit (exitSuccess)
> import System.IO 
> import System.Console.Editline.Readline

> undoC = Command { cmdname = Undo
>                 , action  = undoAction
>                 , longStr = "undo"
>                 , short = "u"
>                 , help  = helpUndo
>                 }

> quitC = Command { cmdname = Quit
>                 , action  = lift . const exitSuccess
>                 , longStr = "quit"
>                 , short = "q"
>                 , help  = "Sale del programa"
>                 }

> helpC = Command { cmdname = Help
>                 , action  = helpAction
>                 , longStr = "help"
>                 , short = "h"
>                 , help  = helpHelp
>                 } 


> showC = Command { cmdname = Show
>                 , action  = showAction
>                 , longStr = "show"
>                 , short = "sh"
>                 , help  = helpShow
>                 } 

> nextC = Command { cmdname = Next
>                 , action  = nextAction
>                 , longStr = "next"
>                 , short = "n"
>                 , help  = helpNext
>                 }

> startC = Command { cmdname = Start
>                  , action  = startAction
>                  , longStr = "start"
>                  , short = "s"
>                  , help  = helpStart
>                  }

> addRuleC = Command { cmdname = AddRule
>                    , action  = addRuleAction
>                    , longStr = "add"
>                    , short = "a"
>                    , help  = helpAddRule
>                    }

> addConstC = Command { cmdname = AddDef
>                     , action  = addConstant
>                     , longStr = "def"
>                     , short = "d"
>                     , help  = helpAddDef
>               }

> commands :: [Command]
> commands = [ quitC
>            , helpC
>            , showC
>            , addRuleC
>            , nextC
>            , startC
>            , undoC
>            , addConstC
>            ]
             

> putStrLift :: Action
> putStrLift = lift . putStrLn 

> putStrPretty :: Pretty a => a -> ChkState ()
> putStrPretty = putStrLift . render .  pretty 

> putStrExpr :: QExpr -> ChkState ()
> putStrExpr = putStrPretty . unlift

> putStrErr :: Action
> putStrErr err = gets derState >>= 
>                 maybe (putStrLift err)
>                       (\der -> maybe (return ())
>                        (putStrLift . ("Error en la línea "++) . show) (lineNo der) >>
>                        (putStrLift err)
>                       )

> emptyDer :: ChkState ()
> emptyDer = gets derState >>= \state -> when (isNothing state) (putStrLift errEmptyDer)


> showDerStep :: DerState -> Axiom -> QExpr -> Substitution -> ChkState ()
> showDerStep der rule res sub = putStrPretty $ mkder (lhs der) [stepsub rule res [sub]]
>     where lhs = conclusion . derivation

> showStep :: Axiom -> QExpr -> ChkState ()
> showStep rule res = putStrPretty $ step rule res


> endDer :: ChkState ()
> endDer = guardState (return ()) (\der -> 
>               putStrPretty (derivation der) >>
>               putStrLift "Ha concluido la derivación." >>
>               maybe (return ()) (putStrLift . ("Líneas leídas: " ++) . show) (lineNo der)
>               )

> showDer :: ChkState ()
> showDer = guardState emptyDer (putStrPretty . derivation)

> showThms :: ChkState ()
> showThms = gets theorems >>= mapM_ putStrLift . showRules 

> startAction :: Action
> startAction params = do exps <- liftM splitParams (return params)
>                         unless (length exps < 2) $ do
>                           [e1,e2] <- return exps
>                           parseExpL e1 (\from -> parseExpL e2 
>                                   (\goal -> setState (mkder from []) (return goal) >> 
>                                            putStrExpr from ))

> nextAction :: Action
> nextAction params = guardState emptyDer $ \der -> do 
>                      parts <- liftM splitParams (return params)
>                      unless (length parts < 2) $ do
>                        [ruleName, expr] <- return parts
>                        parseRuleL ruleName $ \rule -> 
>                          parseExpL expr $ \res -> do 
>                            addStepSt rule res >>= \sub ->
>                                maybe (return ())
>                                      (\s -> maybe (return ())
>                                            (\g -> if res == g then endDer else showDerStep der rule res s)
>                                            (goal der)) sub
>                          

> addRuleAction :: Action
> addRuleAction params = guardState emptyDer (\der -> 
>                       if (null . steps . derivation) der
>                       then putStrErr errNoStep
>                       else do (s,t) <- (return . (lhsDer &&& result . last . steps) . derivation) der
>                               getParams params 2 (\ (ruleName:ruleAbr:_) -> 
>                                                     (return (axiom ruleName (s `equiv` t) [ruleAbr] ruleName) >>=
>                                                      modify . addRule >>
>                                                      putStrLift ("Se agregó la regla "++ ruleName)))
>                                                  (putStrErr errFewParams))

> addConstant :: Action
> addConstant params = getParams params 4 (\(ruleName:ruleAbr:constName:constDef:_) ->
>                                                  parseConstL constName (\_ ->
>                                                  parseExpL constDef (\def ->
>                                                   (return (definition ruleName constName def [ruleAbr] ruleName) >>=
>                                                    modify . addRule >>
>                                                    putStrLift ("Se agregó la constante "++ ruleName)))))
>                                         (putStrErr errFewParams)

> showRule :: Action
> showRule name = parseRuleL name (putStrExpr . prop)

> showAction :: Action
> showAction param = case param of
>                   "rules" -> showThms
>                   "derivation" -> showDer
>                   "goal" -> guardState emptyDer ((maybe (return ()) putStrExpr) . goal)
>                   what -> if take 4 what == "rule" then showRule (drop 5 what)
>                          else putStrLift helpShow

> helpAction :: Action
> helpAction = maybe (putStrLift helpHelp) putStrLift . liftM help . readCmd commands
  
> undoAction :: Action
> undoAction = const $ guardState emptyDer (const undoStepState)

> parseExpL :: String -> (QExpr -> ChkState ()) -> ChkState ()
> parseExpL e next = liftM parseExp (return e) >>= either (putStrLift . show) next 

> parseConstL :: String -> (ConName -> ChkState ()) -> ChkState ()
> parseConstL e next = liftM parseConst (return e) >>= either (putStrLift . show) next 

> parseRuleL :: String -> (Axiom -> ChkState ()) -> ChkState ()                    
> parseRuleL r next = gets theorems >>= 
>                     liftM (flip parseRule r) . return >>=
>                     maybe (putStrErr errNoSuchRule) next


> interaction :: ChkState ()
> interaction = lift (readline "> ") >>= 
>               maybe (return "") return >>= 
>               (\l -> if null l then return () else 
>                     return (parseCommand commands l) >>=
>                     maybe (putStrErr errMissingCmd) (\ m -> addLine l m >>= uncurry ($)) >>
>                     putStrLift "") >>
>               interaction
>     where addLine l m = lift (addHistory l) >> return m

> readLineY :: Handle -> ChkState String
> readLineY handle = liftedIf (lift (hIsEOF handle))
>                    (lift (return ""))
>                    (gets derState >>= 
>                     maybe (return ()) (\d -> modify (\c -> c { derState = return (nextLine d)})) >> 
>                     lift (hGetLine handle) >>= 
>                     \line -> if commented line
>                             then readLineY handle
>                             else return line)
>     where commented :: String -> Bool
>           commented =  (=="--") . take 2

> compFile :: Handle -> ChkState ()
> compFile h = do lineRule <- readLineY h
>                 lineRes  <- readLineY h
>                 liftedIf (return (null lineRule)) 
>                          (endDer >> lift (hClose h) >> lift exitSuccess)
>                          (parseRuleL lineRule 
>                                      (\ax -> parseExpL lineRes (\res -> 
>                                              liftedIf (liftM isNothing (addStepSt ax res))
>                                                       (putStrErr (errBadStep ++ name ax) >> 
>                                                        lift exitSuccess)
>                                                       (compFile h)))
>                          )
                       
> checkHandle :: String -> Handle -> ChkState ()
> checkHandle name handle = do 
>   atEOF <- lift $ hIsEOF handle
>   unless atEOF $ do
>     exprL <- readLineY handle
>     parseExpL exprL (\expr ->
>                      setStateF expr 1 name >>
>                      compFile handle)

> checkFile :: FilePath -> ChkState ()
> checkFile fname = do 
>   handle <- lift $ openFile fname ReadMode
>   checkHandle fname handle

> checkStdIn :: ChkState () 
> checkStdIn = checkHandle "stdin" stdin

> welcome, welcomeInt :: String
> welcome =  "yahc: Versión: 0.3"
> welcomeInt = "Para ver los posibles comandos y lo que hacen podes hacer \"help\""
